package java.awt;
import java.awt.event.*;
import java.io.*;
import java.util.Vector;
import java.lang.reflect.*;
import java.awt.peer.*;

public class Container extends Component {

    // TODO: use collections API instead...
    Vector components = new Vector();
    Insets insets;
    LayoutManager layout;

    /** Events that lightweight children needs. */
    long childrenEventMask;

    public Container() {}

    public int getComponentCount() {
	return components.size();
    }

    /** @deprecated use getComponentCount() */
    public int countComponents() {
	return getComponentCount();
    }

    public Component getComponent(int index) {
	return (Component) components.elementAt(index);
    }

    public Component[] getComponents() {
	throw new UnsupportedOperationException("not implemented");
    }

    /** Note that using the XToolkit, the insets returned by this
        method deviate from the Sun Javadoc description. The Sun
        implementation tries to describe the thickness of the border
        provided by the window manager as insets. This is plain wrong
        behavior in X11. Application should only provide hints to the
        window manager and leave all details of window management and
        decoration to the window manager. Including window decorations
        thickness in insets cannot be done in a reliable way. The X11
        toolkit therefore simply returns [0, 0, 0, 0] as inset. This
        is also correct according to the coordinate space of the
        toplevel windows given by the X11 backend. */
    public Insets getInsets() {
	ContainerPeer peer = (ContainerPeer) getPeer();
	if (peer == null) return new Insets(0, 0, 0, 0);
	
	return peer.getInsets();
    }
    

    /** @deprecated use getInsets() */
    public Insets insets() {
	return getInsets();
    }

    public Component add(Component child) {
	add(child, null);
	return child;
    }

    

    /** Use add(Object, Component) instead. */
    public Component add(String constraintTag, Component child) {
	addImpl(child, constraintTag, -1);
	return child;
    }
    
    public Component add(Component child, int index) {
	addImpl(child, null, index);
	return child;
    }

    public void add(Component child, Object constraints) {
	addImpl(child, constraints, -1);
    }
    
    public void add(Component child, Object constraints, int index) {
	addImpl(child, constraints, index);
    }

    protected void addImpl(Component child, Object constraint, int index) {

	// FIXME: what if child is already owned by another container?

	if (index == -1) {
	    components.addElement(child);
	} else {
	    components.insertElementAt(child, index);
	}

	child.parent = this;
	if (getPeer() != null) {
	    child.addNotify();

	    if (child.isLightweight()) {
		changeChildrenEventMask(childrenEventMask |
					child.getEventMask());
	    }

	}
	

	LayoutManager layout = getLayout();
	if (layout != null) {
	    if (layout instanceof LayoutManager2) {
		((LayoutManager2) layout).addLayoutComponent(child,
							     constraint);
	    } else {
		layout.addLayoutComponent(constraint.toString(), child);
	    }
	    
	}
    }

    void changeChildrenEventMask(long newMask) {
	System.out.println(this + ": changing children event mask");
	long oldCombined = getEventMask();
	childrenEventMask = newMask;
	long newCombined = getEventMask();

	if (oldCombined != newCombined) fireEventMaskChange();
    }

    void lightweightChildEventMaskChanged(Component lightweight) {
	/* FIXME: the current implementation only handles the addition
	   of events, not the removal. */

	System.out.println(this + ": parent notified");

	changeChildrenEventMask(childrenEventMask |
				lightweight.getEventMask());
    }
    

    public void remove(int index) {
	// FIXME: do locking
	Component child = (Component) components.elementAt(index);
	removeImpl(child);
	components.removeElementAt(index);
    }
    
    public void remove(Component child) {
	// FIXME: do locking
	removeImpl(child);
	components.removeElement(child);
    }

    public void removeAll() {
	// FIXME: do locking
	int count = components.size();
	LayoutManager layout = getLayout();
	for (int i=0; i<count; i++) {
	    Component child = (Component) components.elementAt(i);
	    removeImpl(child);
	}
	
	components.removeAllElements();
    }

    void removeImpl(Component child) {
	LayoutManager layout = getLayout();
	if (layout != null) getLayout().removeLayoutComponent(child);
	child.removeNotify();
	child.parent = null;
    }
    


    /** Gets the layout manager that is used by the container to lay
        out its children. A layout manager is not required for a
        layout manager. If no layout manager is present, the bounds of
        the children are not affected by doLayout(). All code using
        this method should be prepared to get a null reference.

	@return the layout manager or null if no layout manager. */
    public LayoutManager getLayout() {
	return layout;
    }

    public void setLayout(LayoutManager manager) {
	layout = manager;
    }

    public void doLayout() {
	LayoutManager layout = getLayout();
	if (layout == null) return;

	layout.layoutContainer(this);
    }

    /** @deprecated use doLayout() */
    public void layout() {
	doLayout();
    }
    
    public void invalidate() {
	isValid = false;

	Container parent = getParent();
	if ((parent != null) && !parent.isValid()) parent.invalidate();
    }

    /** Validates the component. For a container this means that its
	children are layed out. The laying out of components recurses
	down the containment hierarcy, subdividing the available
	space. */


    public void validate() {
	if (isValid()) return;

	// FIXME: do locking
	validateTree();
    }

    protected void validateTree() {
	if (isValid) return;
	doLayout(); // lay out the children of this component

	int count = components.size();
	for (int i=0; i<count; i++) {
	    Component component = (Component) components.elementAt(i);
	    component.validateTree();
	}

	/* children will call invalidate() when they are layed out. It
	   is therefore imporant that isValid is not set to true
	   before after the children has been layed out. */
	isValid = true;
    }
    

    public void setFont(Font font) {
	super.setFont(font);
    }
    
    public Dimension getPreferredSize() {
	LayoutManager layout = getLayout();
	if (layout == null) return super.getPreferredSize();
	
	return getLayout().preferredLayoutSize(this);
    }
    
    /** @deprecated use getPreferredSize() */
    public Dimension preferredSize() {
	return getPreferredSize();
    }

    public Dimension getMinimumSize() {
	LayoutManager layout = getLayout();
	if (layout == null) return super.getMinimumSize();
	
	return layout.minimumLayoutSize(this);
    }

    /** @deprecared use getMinimumSize() */
    public Dimension minimumSize() {
	return getMinimumSize();
    }

    public Dimension getMaximumSize() {
	LayoutManager layout = getLayout();
	if ((layout != null) && (layout instanceof LayoutManager2)) {
	    return ((LayoutManager2) layout).maximumLayoutSize(this);
	}
	
	return super.getMaximumSize();
    }

    public float getAlignmentX() {
	LayoutManager layout = getLayout();
	if ((layout != null) && (layout instanceof LayoutManager2)) {
	    return ((LayoutManager2) layout).getLayoutAlignmentX(this);
	}
	
	return super.getAlignmentX();
    }
    
    public float getAlignmentY() {
	LayoutManager layout = getLayout();
	if ((layout != null) && (layout instanceof LayoutManager2)) {
	    return ((LayoutManager2) layout).getLayoutAlignmentY(this);
	}
	
	return super.getAlignmentX();
    }
    
    public void paint(Graphics gfx) {
	if (!isShowing()) return;
	super.paint(gfx);
	visitChildren(gfx, GfxPaintVisitor.INSTANCE, true);
    }


    void visitChildren(Graphics gfx, GfxVisitor visitor,
			  boolean lightweightOnly) {
	// FIXME: do locking
	int count = components.size();
	for (int i=count-1; i>=0; i--) {
	    Component component = (Component) components.elementAt(i);
	    // ignore invisible children
	    if (!component.isVisible()) continue;

	    visitChild(gfx, visitor, lightweightOnly, component);
	}
    }

    private void visitChild(Graphics gfx, GfxVisitor visitor,
			    boolean lightweightOnly, Component component) {
	boolean isHeavy = !component.isLightweight();
	
	if (lightweightOnly && isHeavy) return;

	Rectangle bounds = component.getBounds();
	Rectangle clip = gfx.getClipBounds().intersection(bounds);
	
	if (clip.isEmpty()) return;
	
	Graphics gfx2 = gfx.create();
	gfx2.setClip(clip);
	gfx2.translate(bounds.x, bounds.y);
	
	visitor.visit(component, gfx2);
    }
    
    public void update(Graphics gfx) {
	super.update(gfx); // not sure what to do differently
    }

    public void print(Graphics gfx) {
	super.print(gfx);
	visitChildren(gfx, GfxPrintVisitor.INSTANCE, true);
    }

    public void paintComponents(Graphics gfx) {
	super.paint(gfx);
	visitChildren(gfx, GfxPaintAllVisitor.INSTANCE, true);
    }

    public void printComponents(Graphics gfx) {
	super.paint(gfx);
	visitChildren(gfx, GfxPrintAllVisitor.INSTANCE, true);
    }

    public void addContainerListener(ContainerListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void removeContainerListener(ContainerListener listener) {
	throw new UnsupportedOperationException("not implemented");
    }
    
    /*
    public EventListener[] getListeners(Class listenerType) {
	throw new UnsupportedOperationException("not implemented");
    }
    */
    
    protected void processEvent(AWTEvent evt) {
	super.processEvent(evt); // why override?
    }

    protected void processContainerEvent(ContainerEvent evt) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void deliverEvent(Event evt) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Component getComponentAt(int x, int y) {
	if (!contains(x, y)) return null;

	int count = components.size();
	for (int i=0; i<count; i++) {
	    Component component = (Component) components.elementAt(i);	

	    // Ignore invisible children...
	    if (!component.isVisible()) continue;

	    Rectangle bounds = component.getBounds();
	    if (bounds.contains(x, y)) return component;
	}
	return this;
    }

    public Component locate(int x, int y) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Component getComponentAt(Point p) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Component findComponentAt(int x, int y) {
	throw new UnsupportedOperationException("not implemented");
    }

    public Component findComponentAt(Point p) {
	throw new UnsupportedOperationException("not implemented");
    }

    public void addNotify() {
	super.addNotify();
    }

    void addNotifyContainerChildren() {

	// FIXME: do locking
	int count = components.size();
	for (int i=0; i<count; i++) {
	    Component child = (Component) components.elementAt(i);
	    child.addNotify();
	    
	    if (child.isLightweight()) {
		childrenEventMask |= child.getEventMask();
	    }
	}
    }

    long getEventMask() {
	return super.getEventMask() | childrenEventMask;
    }

    public void removeNotify() {
	// FIXME: do locking
	int count = components.size();
	for (int i=0; i<count; i++) {
	    Component component = (Component) components.elementAt(i);
	    component.removeNotify();
	}
	super.removeNotify();
    }
    
    public boolean isAncestorOf(Component c) {
	while (c != null) {
	    if (c == this) return true;
	    c = c.getParent();
	}
	return false;
    }

    protected String paramString() 
    {
	StringBuffer sb = new StringBuffer(super.paramString());
	// FIXME: do locking

	/*  // too much clutter?
	int count = components.size();
	for (int i=0; i<count; i++) {
	    Component component = (Component) components.elementAt(i);
	    sb.append(",");
	    sb.append(component);
	}
	*/

	LayoutManager layout = getLayout();
	if (layout != null) sb.append(layout.getClass().getName());
	
	return sb.toString();
    }
    
    public void list(PrintStream out, int indent) {
	throw new UnsupportedOperationException("not implemented");
    }

    
    public void list(PrintWriter out, int indent) {
	throw new UnsupportedOperationException("not implemented");
    }
}



abstract class GfxVisitor {
    public abstract void visit(Component c, Graphics gfx);
}

class GfxPaintVisitor extends GfxVisitor {
    public void visit(Component c, Graphics gfx) { c.paint(gfx); }
    public static final GfxVisitor INSTANCE = new GfxPaintVisitor();
}

class GfxPrintVisitor extends GfxVisitor {
    public void visit(Component c, Graphics gfx) { c.print(gfx); }
    public static final GfxVisitor INSTANCE = new GfxPrintVisitor();
}

class GfxPaintAllVisitor extends GfxVisitor {
    public void visit(Component c, Graphics gfx) { c.paintAll(gfx); }
    public static final GfxVisitor INSTANCE = new GfxPaintAllVisitor();
}

class GfxPrintAllVisitor extends GfxVisitor {
    public void visit(Component c, Graphics gfx) { c.printAll(gfx); }
    public static final GfxVisitor INSTANCE = new GfxPrintAllVisitor();
}


	

    
    
